package main

import (
	"bufio"
	"flag"
	"fmt"
	"math/big"
	"os"
	"os/exec"
	"strconv"
)

type Rand struct {
	s *bufio.Scanner
}

func NewRand(seed int) *Rand {
	cmd := exec.Command("../randhelper", strconv.Itoa(seed))
	stdout, err := cmd.StdoutPipe()
	if err != nil {
		panic(err)
	}
	err = cmd.Start()
	if err != nil {
		panic(err)
	}
	s := bufio.NewScanner(stdout)
	s.Split(bufio.ScanWords)
	return &Rand{s}
}

func (rand *Rand) Next() int {
	if !rand.s.Scan() {
		panic("out of rand")
	}
	n, err := strconv.Atoi(rand.s.Text())
	if err != nil {
		panic(err)
	}
	return n
}

func calcIterationCount(r uint32) int {
	upper := uint32(uint64(int64(r)*int64(-0x7d4996f1)) >> 32)
	return int(r - 0xfab*((r+upper)>>11) - (r >> 31))
}

func main() {
	var one, two big.Int
	one.SetUint64(1)
	two.SetUint64(2)

	flag.Parse()
	seed, err := strconv.Atoi(flag.Arg(0))
	if err != nil {
		panic(err)
	}
	rand := NewRand(seed)

	filename := flag.Arg(1)
	f, err := os.Open(filename)
	if err != nil {
		panic(err)
	}
	defer f.Close()

	iterationCount := calcIterationCount(uint32(rand.Next()))
	rands := make([]int, iterationCount)
	for i := 0; i < iterationCount; i++ {
		rands[i] = rand.Next()
	}

	var cfile []*big.Int
	s := bufio.NewScanner(f)
	s.Split(bufio.ScanWords)
	for s.Scan() {
		n := new(big.Int)
		n.SetString(s.Text(), 10)
		cfile = append(cfile, n)
	}
	if s.Err() != nil {
		panic(s.Err())
	}
	// reverse
	for i := 0; i < len(cfile)/2; i++ {
		tmp := cfile[i]
		cfile[i] = cfile[len(cfile)-1-i]
		cfile[len(cfile)-1-i] = tmp
	}

	size := len(cfile)
	// size := 8

	for i := 0; i < iterationCount; i++ {
		// r := rands[iterationCount - 1 - i]
		// t := (r % len(cfile)) | 1
		// fmt.Fprintf(os.Stderr, "xor %d\n", t)
		// cfile[t].Xor(cfile[t], &one)

		for j := 0; j < size; j += 2 {
			// cfile[j]   == a + b
			// cfile[j+1] == a - b
			// (a + b) + (a - b) == 2 * a
			var a big.Int
			a.Add(cfile[j], cfile[j+1])
			if a.Bit(0) != 0 {
				cfile[j+1].Xor(cfile[j+1], &one)
				a.Add(cfile[j], cfile[j+1])
				// fmt.Fprintf(os.Stderr, "foo %d %d %s %s %s\n", i, j, cfile[j], cfile[j+1], a.String())
			}
			a.Div(&a, &two)
			cfile[j+1].Sub(cfile[j], &a)
			cfile[j].Set(&a)
		}
	}

	buf := make([]byte, size)
	for i := 0; i < size; i++ {
		n := cfile[i]
		if !n.IsUint64() {
			fmt.Printf("%v no\n", seed)
			os.Exit(1)
		}
		u := n.Uint64()
		if u > 255 {
			fmt.Printf("%v no\n", seed)
			os.Exit(1)
		}
		buf[i] = uint8(u)
	}

	// fmt.Printf("%v %+q\n", seed, string(buf))
	os.Stdout.Write(buf)
}
